/** Performs a shallow update of an object using a partial object of the same shape. A new object is returned. */
export declare function update<Obj extends {}, K extends keyof Obj>(host: Obj, spec: Pick<Obj, K>): Obj;
/** Marker used to delete a key */
export declare const DELETE: undefined;
export declare type Leaf = string | number | boolean | null | symbol | Date | Function;
export declare type AtUpdater<TARGET, CURRENT> = [CURRENT] extends [any[]] ? ArrayAtUpdater<TARGET, CURRENT> : ObjectAtUpdater<TARGET, CURRENT>;
export declare type BoundAtUpdater<TARGET, CURRENT> = [CURRENT] extends [any[]] ? ArrayBoundAtUpdater<TARGET, CURRENT> : ObjectBoundAtUpdater<TARGET, CURRENT>;
export declare type Updater<TARGET, CURRENT> = [CURRENT] extends [any[]] ? ArrayUpdater<TARGET, CURRENT> : [CURRENT] extends [Leaf] ? AnySetter<TARGET, CURRENT> : ObjectUpdater<TARGET, CURRENT>;
export declare type BoundUpdater<TARGET, CURRENT> = [CURRENT] extends [any[]] ? ArrayBoundUpdater<TARGET, CURRENT> : [CURRENT] extends [Leaf] ? AnyBoundSetter<TARGET, CURRENT> : ObjectBoundUpdater<TARGET, CURRENT>;
export interface ArrayAtUpdater<TARGET, CURRENT> {
    /**
     * Selects an Array index for update or further at() chaining
     */
    at(index: number): Updater<TARGET, [CURRENT] extends [any[]] ? CURRENT[number & keyof CURRENT] | undefined : never>;
}
export interface ObjectAtUpdater<TARGET, CURRENT> {
    /**
     * Selects this Object key for update or further at() chaining
     */
    at<K extends keyof CURRENT>(key: K): Updater<TARGET, CURRENT[K]>;
}
export interface ArrayBoundAtUpdater<TARGET, CURRENT> {
    /**
     * Selects an Array index for update or further at() chaining
     */
    at(index: number): BoundUpdater<TARGET, [CURRENT] extends [any[]] ? CURRENT[number & keyof CURRENT] | undefined : never>;
}
export interface ObjectBoundAtUpdater<TARGET, CURRENT> {
    /**
     * Selects this Object key for update or further at() chaining
     */
    at<K extends keyof CURRENT>(key: K): BoundUpdater<TARGET, CURRENT[K]>;
}
export interface AnySetter<TARGET, CURRENT> {
    /**
     * Sets the value at the currently selected path.
     */
    set(value: CURRENT): (target: TARGET) => TARGET;
    /**
     * Modifies the value at the specified path. The current value is passed.
     */
    modify(modifier: (value: CURRENT) => CURRENT): (target: TARGET) => TARGET;
}
export interface AnyUpdater<TARGET, CURRENT> extends AnySetter<TARGET, CURRENT> {
    /**
     * Makes the previous nullable chain level 'safe' by using a default value
     */
    withDefault(defaultValue: CURRENT): Updater<TARGET, NonNullable<CURRENT>>;
    /**
     * Aborts the whole update operation if the previous chain level is null or undefined.
     */
    abortIfUndef(): Updater<TARGET, NonNullable<CURRENT>>;
    /**
     * Aborts the whole update operation if the previous chain level doesn't verify a type guard
     */
    abortIfNot<B extends CURRENT>(predicate: (value: CURRENT) => value is B): Updater<TARGET, B>;
    /**
     * Aborts the whole update operation if the previous chain level doesn't verify a predicate
     */
    abortIfNot(predicate: (value: CURRENT) => boolean): Updater<TARGET, CURRENT>;
}
export interface ArrayUpdater<TARGET, CURRENT> extends AnyUpdater<TARGET, CURRENT>, ArrayAtUpdater<TARGET, CURRENT> {
}
export interface ObjectUpdater<TARGET, CURRENT> extends AnyUpdater<TARGET, CURRENT>, ObjectAtUpdater<TARGET, CURRENT> {
}
export interface AnyBoundSetter<TARGET, CURRENT> {
    /**
     * Sets the value at the currently selected path.
     */
    set(value: CURRENT): TARGET;
    /**
     * Modifies the value at the specified path. The current value is passed.
     */
    modify(modifier: (value: CURRENT) => CURRENT): TARGET;
}
export interface AnyBoundUpdater<TARGET, CURRENT> extends AnyBoundSetter<TARGET, CURRENT> {
    /**
     * Makes the previous nullable chain level 'safe' by using a default value
     */
    withDefault(defaultValue: CURRENT): BoundUpdater<TARGET, NonNullable<CURRENT>>;
    /**
     * Aborts the whole update operation if the previous chain level is null or undefined.
     */
    abortIfUndef(): BoundUpdater<TARGET, NonNullable<CURRENT>>;
    /**
     * Aborts the whole update operation if the previous chain level doesn't verify a type guard
     */
    abortIfNot<B extends CURRENT>(predicate: (value: CURRENT) => value is B): BoundUpdater<TARGET, B>;
    /**
     * Aborts the whole update operation if the previous chain level doesn't verify a predicate
     */
    abortIfNot(predicate: (value: CURRENT) => boolean): BoundUpdater<TARGET, CURRENT>;
}
export interface ArrayBoundUpdater<TARGET, CURRENT> extends AnyBoundUpdater<TARGET, CURRENT>, ArrayBoundAtUpdater<TARGET, CURRENT> {
}
export interface ObjectBoundUpdater<TARGET, CURRENT> extends AnyBoundUpdater<TARGET, CURRENT>, ObjectBoundAtUpdater<TARGET, CURRENT> {
}
export declare function deepUpdate<TARGET extends object>(target: TARGET): BoundAtUpdater<TARGET, TARGET>;
export declare function deepUpdate<TARGET extends object>(): AtUpdater<TARGET, TARGET>;
